"""
Cache the details of non-words.
"""

import datetime
import hashlib
import io
import logging
import os
import pathlib
import zipfile

import dataset
import requests


class MemoryCache:  # pylint: disable=too-few-public-methods
    """
    Class used to keep lazy memory cache of non-words.
    """

    cache = None
    prepared = None


def get_config_dir(basepath=None):
    """
    Return the directory to store the cache of PyPi packages
    """
    if basepath is None:
        try:
            basepath = pathlib.Path.home()
        except AttributeError:
            basepath = pathlib.Path(os.path.expanduser("~"))
    path = basepath / ".unanimous"
    if not path.exists():
        path.mkdir()
    return path


def get_storage_table(basepath=None):
    """
    Connect to cache db
    """
    path = get_config_dir(basepath=basepath)
    con = dataset.connect("sqlite:///%s" % (path / "cache.db"))
    table = con["storage"]
    if not MemoryCache.prepared:
        prepare_table(con, table)
        MemoryCache.prepared = True
    return table


def prepare_table(con, table):
    """
    Ensure the table and index are setup
    """
    table.create_column("key", con.types.string(25))
    table.create_column("value", con.types.text)
    table.create_index(["key"])


def load_key(key, deflt=None, basepath=None):
    """
    Retrieve stored key value.
    """
    table = get_storage_table(basepath=basepath)
    row = table.find_one(key=key)
    if not row:
        return deflt
    return row["value"]


def save_key_value(key, value, basepath=None):
    """
    Save key, value to storage.
    """
    table = get_storage_table(basepath=basepath)
    table.upsert({"key": key, "value": value}, keys=["key"], ensure=False)


def get_cached_sha(basepath=None):
    """
    Lookup current hash value
    """
    return load_key("sha", basepath=basepath)


def check_upstream_zip_hash(basepath=None):
    """
    Quick check to see if the upstream master.zip has been updated.
    """
    cache_sha = get_cached_sha(basepath=basepath)
    if not cache_sha:
        return False
    url = (
        "https://github.com/resplendent-dev/unanimous"
        "/blob/master/master.sha256?raw=true"
    )
    try:
        response = requests.get(url, timeout=1)
    except OSError:
        logging.exception("Unable to check non-word cache at this time.")
        # Can not be reached - assume not updated
        return False
    else:
        current_sha = response.text.strip().split(" ", 1)[0]
        return cache_sha == current_sha


def get_current_non_words(basepath=None):
    """
    Work out if our cache is up to date and update it if not and then return
    the list of current non-words
    """
    cached_words = get_cached_words(basepath=basepath)
    if cached_words:
        return cached_words
    return update_cached_nonwords(basepath=basepath)


def update_cached_nonwords(basepath=None):
    """
    Load the latest non-words and update the db
    """
    url = (
        "https://github.com/resplendent-dev/unanimous"
        "/blob/master/master.zip?raw=true"
    )
    try:
        response = requests.get(url, timeout=1)
    except OSError:
        logging.exception("Unable to update non-word cache at this time.")
        return force_get_cached_words(basepath=basepath, deflt=set())
    else:
        content = response.content
        return update_cache_with_data(content, basepath=basepath)


def update_cache_with_data(bytedata, basepath=None):
    """
    Update the cache with the provided zip data
    """
    sha256 = hashlib.sha256(bytedata).hexdigest()
    with io.BytesIO(bytedata) as fobj:
        with zipfile.ZipFile(fobj) as zobj:
            data = zobj.read("nonwords.txt").decode("utf-8")
    save_key_value("nonwords", data, basepath=basepath)
    save_key_value("sha", sha256, basepath=basepath)
    save_key_value(
        "timestamp", datetime.datetime.now().strftime("%Y%m%d%H%M%S"), basepath=basepath
    )
    return set(data.splitlines())


def get_cached_words(basepath=None):
    """
    If the cache is current then return the list of words otherwise return
    None
    """
    if MemoryCache.cache:
        return MemoryCache.cache
    timestamp = load_key("timestamp", basepath=basepath)
    if not timestamp:
        return None
    cache_time = datetime.datetime.strptime(timestamp, "%Y%m%d%H%M%S")
    if cache_time + datetime.timedelta(days=1) < datetime.datetime.now():
        # Is cache hash still okay anyway?
        if not check_upstream_zip_hash(basepath=basepath):
            return None
        # Refresh timestamp
        save_key_value(
            "timestamp",
            datetime.datetime.now().strftime("%Y%m%d%H%M%S"),
            basepath=basepath,
        )
    MemoryCache.cache = force_get_cached_words(basepath=basepath)
    return MemoryCache.cache


def force_get_cached_words(basepath=None, deflt=None):
    """
    Grab whatever is in the cache
    """
    nonwords = load_key("nonwords", basepath=basepath)
    if not nonwords:
        return deflt
    return set(nonwords.splitlines())



"""
Support automatically locating a custom non-word list
assumed to be `spelling_wordlist.txt` or `doc(s)/spelling_wordlist.txt`
in any ancestor directory
"""

import io
import pathlib


def get_custom_wordlist(path):
    """
    If found load the custom word list
    """
    result = set()
    if not path:
        return result
    with io.open(path, "r", encoding="utf-8") as fobj:
        for line in fobj:
            result.add(line.lower().strip())
    return result


def locate_custom_wordlist():
    """
    From the current working directory locate the custom word list.
    """
    start = pathlib.Path(".").resolve()
    return locate_custom_wordlist_from(start)


def locate_custom_wordlist_from(path):
    """
    From the provided directory locate the custom word list.
    """
    check = [
        path / "spelling_wordlist.txt",
        path / "docs" / "spelling_wordlist.txt",
        path / "doc" / "spelling_wordlist.txt",
    ]
    for checkpath in check:
        if checkpath.is_file():
            return checkpath
    if path.parent == path:
        return None
    return locate_custom_wordlist_from(path.parent)



"""
pyspelling filter to eliminate words which are not in a dictionary but are not
a spelling mistake but rather are common invented words like PyPi.
"""
import codecs
import re

from pyspelling import filters
from wcmatch import glob

from unanimous.custom_nonwords import get_custom_wordlist
from unanimous.util import is_nonword


class NonWordFilter(filters.Filter):
    """Remove non-words from source"""

    def __init__(self, options, **kwargs):
        super().__init__(options, **kwargs)
        self.non_words = set()
        for target in self.config.get("wordlists", []):
            for match in glob.iglob(
                target, flags=glob.N | glob.B | glob.G | glob.S | glob.O
            ):
                self.non_words.update(get_custom_wordlist(match))

    @staticmethod
    def get_default_config():
        """Get default configuration."""
        return {
            "too_short": 3,
            "wordlists": [],
            "lowercase_only": True,
            "exclude_apostrophe": True,
        }

    def filter(self, source_file, encoding):  # noqa A001
        """Parse text file."""

        with codecs.open(source_file, "r", encoding=encoding) as fobj:
            text = fobj.read()
        return [filters.SourceText(self._filter(text), source_file, encoding, "text")]

    def _is_nonword(self, word):
        """
        Check if a word matches the non-word filter of being either too short
        or a known non-word.
        """
        return is_nonword(
            word=word,
            lowercase_only=self.config["lowercase_only"],
            too_short_check=self.config["too_short"],
            exclude_apostrophe=self.config["exclude_apostrophe"],
            extra_non_words=self.non_words,
        )

    def _filter(self, text):
        """Filter text"""
        words = re.findall("[A-Za-z']+", text)
        result = []
        for word in words:
            if self._is_nonword(word):
                continue
            result.append(word)
        return "\n".join(result)

    def sfilter(self, source):
        """Filter."""

        return [
            filters.SourceText(
                self._filter(source.text), source.context, source.encoding, "text"
            )
        ]


def get_plugin():
    """Return the filter."""

    return NonWordFilter



"""
Load the list of packages from `PyPi`.
"""

from defusedxml.xmlrpc import monkey_patch, unmonkey_patch, xmlrpc_client


def get_package_list():
    """
    Load the list of packages from `PyPi`.
    """
    monkey_patch()
    client = xmlrpc_client.ServerProxy("https://pypi.python.org/pypi")
    packages = client.list_packages()
    unmonkey_patch()
    return packages



#!/usr/bin/env python
"""
Module load handler for execution via:

python -m unanimous
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
import io
import pathlib
import re
import shutil
import sys
import tempfile
import zipfile

import click

from unanimous.pypi_load import get_package_list

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

__version__ = "0.1"


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.version_option(version=__version__)
@click.pass_context
def main(ctxt):
    """
    Main click group handler
    """
    if ctxt.invoked_subcommand is None:
        run_invocation()


@main.command()
def invoke():
    """
    Primary command handler
    """
    run_invocation()


def get_project_path():
    """
    Locate the directory of the root folder for the project
    """
    this_file = pathlib.Path(sys.modules[__name__].__file__).resolve()
    test_dir = this_file.parent
    app_dir = test_dir.parent
    project_dir = app_dir.parent
    return project_dir


def run_invocation():
    """
    Execute the invocation
    """
    basedir = get_project_path()
    zippath = basedir / "master.zip"
    tmpdir = tempfile.mkdtemp()
    tmppath = pathlib.Path(tmpdir)
    tmpnonwordpath = build_nonwords_file(tmppath)
    with zipfile.ZipFile(str(zippath), "w") as zobj:
        zobj.write(tmpnonwordpath, "nonwords.txt")
        zobj.close()
    checksum = sha256(zippath)
    shapath = basedir / "master.sha256"
    with io.open(str(shapath), "w", encoding="utf-8") as fobj:
        print(checksum, file=fobj)


def build_nonwords_file(tmppath):
    """
    Prepare the zip output.
    """
    packages = get_package_list()
    nonwords = set()
    packagewords = set()
    for package in packages:
        for wordpart in re.findall("[a-z]+", package.lower()):
            if len(wordpart) <= 3:
                continue
            packagewords.add(wordpart)
    basedir = get_project_path()
    nonwordpath = basedir / "nonwords.txt"
    tmpnonwordpath = tmppath / "nonwords.txt"
    shutil.copy(str(nonwordpath), str(tmpnonwordpath))
    with io.open(str(tmpnonwordpath), "w", encoding="utf-8") as fobjout:
        with io.open(str(nonwordpath), "r", encoding="utf-8") as fobjin:
            for line in fobjin:
                nonword = line.strip().lower()
                if nonword not in packagewords:
                    nonwords.add(nonword)
                    packagewords.add(nonword)
        for nonword in sorted(list(nonwords)):
            print(nonword, file=fobjout)
    shutil.copy(tmpnonwordpath, nonwordpath)
    with io.open(str(tmpnonwordpath), "w", encoding="utf-8") as fobjout:
        for nonword in sorted(list(packagewords)):
            print(nonword, file=fobjout)
    return tmpnonwordpath


def sha256(filepath):
    """
    Get quick hash of file
    """
    sha = hashlib.sha256()
    with open(str(filepath), "rb") as fobj:
        while True:
            block = fobj.read(4096)
            if not block:
                break
            sha.update(block)
    return sha.hexdigest()


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter
# vim: set ft=python:



"""
Exposed utility calls
"""

from unanimous.store import get_current_non_words


class NonWords:  # pylint: disable=too-few-public-methods
    """
    Cache of non-words
    """

    non_words = set()


def is_nonword(
    word,
    lowercase_only=True,
    too_short_check=3,
    exclude_apostrophe=True,
    extra_non_words=None,
):
    """
    Utility call to check a word is a non-word
    """
    if lowercase_only and not word.islower():
        return True
    if exclude_apostrophe and "'" in word:
        return True
    too_short_result = len(word) <= too_short_check
    if too_short_result:
        return True
    lword = word.lower()
    if not NonWords.non_words:
        NonWords.non_words = set(get_current_non_words())
    non_word_check = lword in NonWords.non_words
    if non_word_check:
        return True
    if extra_non_words is not None:
        non_word_check = lword in extra_non_words
        if non_word_check:
            return True
    return False
